/*
 * Copyright (C) 2001-2011 OSS Nokalva.  All rights reserved.
 *
 */

/*
 * THIS FILE IS PROPRIETARY MATERIAL OF OSS NOKALVA, INC. AND MAY BE USED
 * ONLY BY DIRECT LICENSEES OF OSS NOKALVA, INC. THIS FILE MAY NOT BE
 * DISTRIBUTED.
 */

/*
 * This test shows how you can produce mixed INDEFINITE/DEFINITE encodings
 * with fragmented string components using the DeferDecoding ASN.1 compiler
 * directive.  PDU components, a and c, of type Foo and Bar are defined with
 * the DeferDecoding directive and treated as OpenType values.  The keyword
 * IMPLICIT is used for illustration (see didfrag.asn).  The resulting tag
 * sequences are:
 *
 * a [2] [0] [1] [2] IMPLICIT OCTET STRING, i.e. [2] overlays [UNIVERSAL 4]
 * ...
 * c [4] IMPLICIT SEQUENCE, i.e. [4] overlays [UNIVERSAL 16]
 */


#include "iaapi.h"
#include "didfrag.h"


static int encodeDefiniteLength(unsigned long length, unsigned char *buffer);
static int fragmentBERPrimitive(OssGlobal *world,
			OssBuf *encodedData, int fl, int def, unsigned char utag);


static unsigned char _v1[8] = {
    0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0xFF
};
static Foo _v0 = {
        8,
        _v1
    };
static char _v3[] = "Test";
static Bar _v2 = {
        _v3,
        TRUE
    };
static S s1 = {
    { 1, 0, NULL, &_v0},
    TRUE,
    { 2, 0, NULL, &_v2}
};


int main(void)
{

    int		pdu_num, err = 0;
    OssBuf	encodedData;
    OssGlobal	w, *world = &w;
    void	*decodedData, *buf;


    err = ossinit(world, didfrag);
    if (err) {
	printf("Initialization failed with error code = %d\n", err);
	return err;
    }

    ossPrint(world, "\tEncoding PDU Bar in DEFINITE-length form...\n\n");
    ossSetEncodingFlags(world, DEBUGPDU|DEFINITE);
    ossSetDecodingFlags(world, DEBUGPDU);
    encodedData.value = NULL;
    encodedData.length = 0;
    err = ossEncode(world, Bar_PDU, &_v2, &encodedData);
    if (err) {
	ossPrint(world, "%s\n", ossGetErrMsg(world));
	ossPrint(world, "Error code: %d\n", err);
	ossterm(world);
	return err;
    }
    ossPrintHex(world, (char *)encodedData.value, encodedData.length);

    s1.c.encoded = encodedData.value;
    s1.c.length = encodedData.length;

    ossPrint(world, "\n\tEncoding PDU Foo in INDEFINITE-length form...\n\n");
    ossSetEncodingFlags(world, DEBUGPDU|INDEFINITE);
    encodedData.value = NULL;
    encodedData.length = 0;
    err = ossEncode(world, Foo_PDU, &_v0, &encodedData);
    if (err) {
	ossPrint(world, "%s\n", ossGetErrMsg(world));
	ossPrint(world, "Error code: %d\n", err);
	ossterm(world);
	return err;
    }
    s1.a.encoded = encodedData.value;
    s1.a.length = encodedData.length;

    ossPrint(world, "\n\tEncoded Foo PDU...\n\n");
    ossPrintHex(world, (char *)encodedData.value, encodedData.length);

    ossPrint(world, "\nFragmenting encoded Foo PDU...\n");
    err = fragmentBERPrimitive(world, &encodedData, 1,
				DEFINITE & ossGetEncodingRules(world), 4);

    ossPrint(world, "\n\tFragmented encoded Foo PDU...\n\n");
    ossPrintHex(world, (char *)encodedData.value, encodedData.length);

    s1.a.encoded = encodedData.value;
    s1.a.length = encodedData.length;

    ossPrint(world, "\n\tUnencoded parent PDU S that includes PDU Bar & PDU Foo...\n\n");
    ossSetEncodingFlags(world, DEBUGPDU|INDEFINITE);
    ossSetDecodingFlags(world, DEBUGPDU);
    ossPrintPDU(world, S_PDU, &s1);
    ossPrint(world, "\n\tEncoding PDU S in INDEFINITE-length form...\n\n");
    encodedData.value = NULL;
    encodedData.length = 0;
    err = ossEncode(world, S_PDU, &s1, &encodedData);
    if (err) {
	ossPrint(world, "%s\n", ossGetErrMsg(world));
	ossPrint(world, "Error code: %d\n", err);
	ossFreeBuf(world, s1.a.encoded);
	ossFreeBuf(world, s1.c.encoded);
	ossterm(world);
	return err;
    }
    else {
	ossPrintHex(world, (char *)encodedData.value, encodedData.length);
	ossPrint(world, "\n\tDecoding parent PDU S...\n\n");
	buf  = encodedData.value;
	decodedData = NULL;
	pdu_num = S_PDU;
	err = ossDecode(world, &pdu_num, &encodedData, &decodedData);
	if (err) {
	    ossPrint(world, "%s\n", ossGetErrMsg(world));
	    ossPrint(world, "Error code: %d\n", err);
	    ossFreeBuf(world, buf);
	    ossterm(world);
	    return err;
	}
	else {
	    S *decodedDataSaved = decodedData;

	    ossPrint(world, "\n\tDecoded parent PDU S...\n\n");
	    ossPrintPDU(world, pdu_num, decodedData);

	    pdu_num = Bar_PDU;
	    encodedData.value = decodedDataSaved->c.encoded;
	    encodedData.length = decodedDataSaved->c.length;
	    decodedData = NULL;
	    ossPrint(world, "\n\tDecoding PDU Bar extracted from parent PDU S...\n\n");
	    err = ossDecode(world, &pdu_num, &encodedData, &decodedData);
	    if (err) {
		ossPrint(world, "%s\n", ossGetErrMsg(world));
		ossPrint(world, "Error code: %d\n", err);
	    }
	    else {
		ossPrint(world, "\n\tDecoded PDU Bar...\n\n");
		ossPrintPDU(world, pdu_num, decodedData);
		ossFreePDU(world, pdu_num, decodedData);
	    }

	    pdu_num = Foo_PDU;
	    encodedData.value = decodedDataSaved->a.encoded;
	    encodedData.length = decodedDataSaved->a.length;
	    decodedData = NULL;
	    ossPrint(world, "\n\tDecoding PDU Foo extracted from parent PDU S...\n\n");
	    err = ossDecode(world, &pdu_num, &encodedData, &decodedData);
	    if (err) {
		ossPrint(world, "%s\n", ossGetErrMsg(world));
		ossPrint(world, "Error code: %d\n", err);
	    }
	    else {
		ossPrint(world, "\n\tDecoded PDU Foo...\n\n");
		ossPrintPDU(world, pdu_num, decodedData);
		ossFreePDU(world, pdu_num, decodedData);
	    }

	    ossFreePDU(world, S_PDU, decodedDataSaved);
	    ossFreeBuf(world, buf);
	    ossFreeBuf(world, s1.c.encoded);
	    ossFreeBuf(world, s1.a.encoded);
	    ossterm(world);
	    return 0;
	}
    }
}


/*
 * Function:            encodes a length using the shoert or long form
 *                      as needed.  The encoded length is stored in the
 *                      caller's provided buffer pointed by 'buffer'.
 *
 * Parameters:
 *
 *   "length"       ... length value to be represented in the long form of encoding
 *   "buffer"       ... pointer to buffer to store the result
 *
 * Return value:        number of bytes written to the buffer
 */
static int encodeDefiniteLength(unsigned long length, unsigned char *buffer)
{

    long i, l = length;


    for (i = 0; l; i++)
	l >>= 8;

    i = length <= 127? 1: i + 1;

    for (l = i; i > 0; i--) {
	buffer[i - 1] = (unsigned char)(length & 0x000000FF);
	length >>= 8;
    }

	/*
	 * The very first octet is 1nnn nnnn, where 7 lower bits is the
	 * number of the subsequent octets.
	 */
    if (l > 1)
	*buffer = (unsigned char)(0x80 | l - 1);
    
    return l;
}


/*
 * Function:            takes DEFINITE and INDEFINIRE BER-encoded string
 *                      values and produces fragmented encodings instead,
 *                      writing the fragmented encoding from the end of
 *                      the buffer to its front.
 *
 * Parameters:
 *
 *   "world"        ... pointer to the OSS's global structure
 *   "encodedData"  ... pointer to instance of the OssBuf structure of
 *                      the encoded string to be fragmented
 *   "flen"         ... fragment length
 *   "def"          ... boolean indicating DEFINITE/INDEFINITE form of
 *                      the string encoding
 *   "utag"         ... a UNIVERSAL tag of the string to be fragmented
 *
 * Return value:        number of bytes written to the buffer
 */
static int fragmentBERPrimitive(OssGlobal *world,
			OssBuf *encodedData, int flen, int def, unsigned char utag)
{

    unsigned char *buf, *curEnc, *nestedEnc;
    unsigned char elen[8], *el = elen;
    unsigned char *pdu;
    int l, not, fl, npos, opos, pl, primlen, tlen, primtaglen;
    long nestedRemainingLength, offset, sizefactor = 4;


    nestedRemainingLength = encodedData->length;
    curEnc = encodedData->value;
    offset = 0;
    for (not = 0; ; not++) {
		/*
		 * Parse a constructed tag-length pair.
		 */
	if ((nestedEnc = ossGetNestedBEREncoding(world, curEnc,
				&nestedRemainingLength)) && nestedEnc) {
		/*
		 * Update "offset" and other variables.
		 */
	    offset += nestedEnc - curEnc;
	    nestedRemainingLength -= nestedEnc - curEnc;
	    curEnc += nestedEnc - curEnc;
	    continue;
	}
	else {
	    primlen = ossEncodingLength(world, curEnc, nestedRemainingLength);
	    primtaglen = ossEncodingIdentifierLength(world, curEnc, nestedRemainingLength);
	    break;
	}
    }

    if (def)
	not = 0;

    pdu = malloc(sizefactor * encodedData->length);
    npos = sizefactor * encodedData->length - 1;
    buf = encodedData->value;

    npos -= not * 2;
    memset(&pdu[npos], 0, not * 2);

    opos = encodedData->length - not * 2;
    pl = primlen;
    tlen = 0;
    while (pl > 0) {
	for (fl = flen; fl; fl--, pl--, tlen++)
	    pdu[--npos] = buf[--opos];
	l = encodeDefiniteLength(flen, elen);
	npos -= l;
	memcpy(&pdu[npos], elen, l);
	tlen += l;
	memcpy(&pdu[--npos], &utag, 1);
	tlen++;
	if (pl < flen)
	    flen = pl;
    }

    l = encodeDefiniteLength(tlen, elen);
    tlen += l;
    npos -= l;
    memcpy(&pdu[npos--], elen, l);

    *curEnc |= 0x20;		/* set the constructed bit to 1 */
    memcpy(&pdu[npos], curEnc, primtaglen);

    npos -= offset;

    memcpy(&pdu[npos], encodedData->value, offset);

    ossFreeBuf(world, encodedData->value);
    encodedData->value = &pdu[npos];
    encodedData->length = sizefactor * encodedData->length - 1 - npos;
    buf = malloc(encodedData->length);
    memcpy(buf, encodedData->value, encodedData->length);
    free(pdu);
    encodedData->value = buf;
    return 0;
}

